/*
 * Copyright (C) 2013 NAlek.
 * Web: http://www.nalek.org
 * E-Mail: nalek@nalek.org
 *
 * ICMPHelper version 1.1
 * Date: 31.03.2011
 *
 */

#include "stdafx.h"
#include "ICMPHelper.h"

//
// Query Perfprmance Timer
//

//     
__int64 CQPTimer::Get()
{
	__int64 i64Ret =0;
	HANDLE hThr =::GetCurrentThread();

	DWORD_PTR dwMask = ::SetThreadAffinityMask( hThr, 0 );

	//    
	::QueryPerformanceCounter( (LARGE_INTEGER *) &i64Ret );

	::SetThreadAffinityMask( hThr,  dwMask);

	return i64Ret;
}

//   
double CQPTimer::Elapsed( __int64 i64Begin, __int64 i64End )
{
	__int64 i64Freq =0;

	::QueryPerformanceFrequency( (LARGE_INTEGER *) &i64Freq );

	return i64Freq ==0 ? 0.0 : ((double)( i64End-i64Begin )/(double)i64Freq)*1000.0;
}


//
// ICMP helper
//

//
//   
//

#ifdef ICMPHLP_WSINIT
//
//  WinSock  
//
bool CICMPHelper::InitWinSock( WORD wVersion )
{
	bool bRet =true;

	WSADATA wsaData;
	if( WSAStartup( wVersion, &wsaData ) )
	{
		// ERROR
		// load winsock library
		bRet =false;   
    }
	else
		if( LOBYTE(wsaData.wVersion) !=LOBYTE(wVersion) || HIBYTE(wsaData.wVersion) !=HIBYTE(wVersion) )
		{
			// ERROR
			// wrong version
			WSACleanup();
			bRet =false;
		}

	return bRet;
}

bool CICMPHelper::CloseWinSock()
{
	return WSACleanup() ==0 ? true : false;
}
#endif

bool CICMPHelper::OpenHost( const char *pHostName )
{
	bool bRet =true;
	hostent *pHostEntry;

	memset( &saiDstHost, 0, sizeof(saiDstHost) );

	//
	//  
	//
	//    IPv4 
	if( inet_addr ( pHostName ) !=INADDR_NONE )
	{
		//  ,    
		//   inet_addr
		saiDstHost.sin_addr.s_addr = inet_addr( pHostName );
		saiDstHost.sin_family =AF_INET;
	}
	else
		//     IPv4 , ,    
		//   
		if( ( pHostEntry =gethostbyname( pHostName ) ) !=NULL )
		{
			//   IPv4  
			//saiHost.sin_addr =*((LPIN_ADDR) *lpHostEntry->h_addr_list);
			memcpy( &saiDstHost.sin_addr, pHostEntry->h_addr, pHostEntry->h_length );
			saiDstHost.sin_family =pHostEntry->h_addrtype;
		}
		else
		{
			// ERROR
			//    
			//::WSASetLastError( WSAHOST_NOT_FOUND );
			bRet =false;
		}

	return bRet;
}

//
//  
//
bool CICMPHelper::OpenICMPSession( int iRecvTimeout, int iSendTimeout, int iBufferSize, bool bBind )
{
	bool bRet =true;

	_iRecvTimeout =iRecvTimeout;
	
	// create socket
	if( ( sSock =socket( AF_INET, SOCK_RAW, IPPROTO_ICMP ) ) ==SOCKET_ERROR )
	{
		// ERROR
		// create raw socket
		bRet =false;
	}
	else
	{
		//  
		//   
		if( bBind )
		{
			// trying to bind
			sockaddr_in saLocalHost ={NULL};
			saLocalHost.sin_family =AF_INET;
			saLocalHost.sin_port =0;
			saLocalHost.sin_addr.s_addr =INADDR_ANY;
			if( bind( sSock, (struct sockaddr*)&saLocalHost, sizeof(saLocalHost) ) ==SOCKET_ERROR )
			{
				// ERROR
				// bind error
				bRet =false;
			}

		}// if


		if( bRet )
		{
			//  
			if( ( setsockopt( sSock, SOL_SOCKET, SO_RCVTIMEO, (char*)&iRecvTimeout, sizeof(int) ) ==SOCKET_ERROR ) ||
				( setsockopt( sSock, SOL_SOCKET, SO_SNDTIMEO, (char*)&iSendTimeout, sizeof(int) ) ==SOCKET_ERROR ) )
			{
				// ERROR
				bRet =false;
			}
			else
			{
				//      MAX_ICMP_DATA_SIZE
				iICMPDataSize =iBufferSize >MAX_ICMP_DATA_SIZE ? MAX_ICMP_DATA_SIZE : iBufferSize ;
				iICMPPacketSize =ICMP_HEADER_LEN+iICMPDataSize;
			}
		}
	}

	return bRet;
}

//
//  
//
bool CICMPHelper::CloseICMPSession()
{
	bool bRet =ICMPHLP_ERR_SUCCESS;

	if( closesocket( sSock ) ==SOCKET_ERROR )
	{
		// ERROR
		// close socket
		bRet =false;
	}

	return bRet;
}

//
//    
//
unsigned short CICMPHelper::IPChecksum( unsigned short *usBuffer, int iSize )
{
	unsigned long cksum = 0;

	// Sum all the words together, adding the final byte if size is odd
	while( iSize >1 )
	{
		cksum +=*usBuffer++;
		iSize -=sizeof( unsigned short );
	}

	if( iSize )
	{
		cksum += *(unsigned char*)usBuffer;
	}

	// Do a little shuffling
	cksum = (cksum >> 16) + (cksum & 0xffff);
	cksum += (cksum >> 16);

	// Return the bitwise complement of the resulting mishmash
	return (unsigned short)(~cksum);
}

//
//  ICMP 
//
int CICMPHelper::SendEchoRequest( unsigned short usPacketID, unsigned short usSeq, int &iSendBytes )
{
	int iRet =ICMPHLP_ERR_SUCCESS;

	//  ICMP 
	//int iPacketSize =ICMP_HEADER_LEN+iICMPDataSize;
	ICMPHeader *sICMPHdr =(ICMPHeader *)malloc( iICMPPacketSize );

	if( sICMPHdr ==NULL )
	{
		// ERROR
		// memory allocation
		//::WSASetLastError( WSA_NOT_ENOUGH_MEMORY );
		iRet =ICMPHLP_ERR_MEMALLOC;
	}
	else
	{
		//   ICMP 
		sICMPHdr->ucType =ICMP_ECHO;
		sICMPHdr->ucCode =0;
		sICMPHdr->usChecksum =0;
		sICMPHdr->usID =usPacketID;
		sICMPHdr->usSeq =usSeq;
		sICMPHdr->ulIDEx =::GetCurrentThreadId();
		sICMPHdr->i64Timestamp =0;

		//  ICMP  
		memset( (char *)sICMPHdr+ICMP_HEADER_LEN, 0xFF, iICMPDataSize );

		//       
		sICMPHdr->i64Timestamp =cTimer.Get();
		sICMPHdr->usChecksum =IPChecksum( (unsigned short *)sICMPHdr, iICMPPacketSize );
		
		iSendBytes =0;
		int iSent =0, iSendSize =iICMPPacketSize;
		while(1)
		{
			//  ICMP 
			iSent =sendto( sSock, ((char *)sICMPHdr)+iSendBytes, iSendSize, 0, (sockaddr*)&saiDstHost, sizeof(saiDstHost));

			if( iSent ==SOCKET_ERROR )
			{
				if( ::WSAGetLastError() ==WSAETIMEDOUT )
				{
					//   
					iRet =ICMPHLP_ERR_TIMEOUT;
					break;
				}
				else
				{
					// ERROR
					//  WinSock
					iRet =ICMPHLP_ERR_WINSOCK;
					break;
				}
			}
			else
				if( iSent <iSendSize )
				{
					//    
					//  
					iSendBytes +=iSent;
					iSendSize =iICMPPacketSize -iSendBytes;
				}
				else
				{
					// ICMP   
					iSendBytes +=iSent+IPv4_HEADER_LEN;
					break;
				}
		}

		free(sICMPHdr);
	}
	return iRet;
}

//
//  ICMP 
//
int CICMPHelper::ReceiveEchoReply( unsigned short usPacketID, unsigned short usSeq, sockaddr_in &saiSrcHost, int &iRecvBytes, double &dRecvTime, int &iRecvTTL )
{
	int iRet =ICMPHLP_ERR_SUCCESS;

	//    
	//    =  IP  +  ICMP  +   

	int iPacketSize =MAX_PACKET_SIZE;
	IPv4Header *sIPHdr =(IPv4Header *)malloc( iPacketSize );

	CQPTimer cAdvTimer;
	if( sIPHdr ==NULL )
	{
		// ERROR
		// memory allocation
		//::WSASetLastError( WSA_NOT_ENOUGH_MEMORY );
		iRet =ICMPHLP_ERR_MEMALLOC;
	}
	else
	{
		__int64 iBegin =cAdvTimer.Get();
		while( 1 )
		{
			int iSrcLen =sizeof(saiSrcHost);
			//  ICMP 

			int iRecv =recvfrom( sSock, (char *)sIPHdr, iPacketSize, 0, (sockaddr *)&saiSrcHost, &iSrcLen );

			if( iRecv ==SOCKET_ERROR )
			{
				if( ::WSAGetLastError() ==WSAETIMEDOUT )
				{
					//   
					iRet =ICMPHLP_ERR_TIMEOUT;
					break;
				}
				else
				{
					// ERROR
					//  WinSock
					iRet =ICMPHLP_ERR_WINSOCK;
					break;
				}
			}
			else
			{
				//    ICMP      sizeof(IPv4Header)+sizeof(ICMPHeader)
				if( iRecv <IPv4_HEADER_LEN+ICMP_HEADER_LEN )
				{
					if( cAdvTimer.Elapsed( iBegin, cAdvTimer.Get() ) >=(double)_iRecvTimeout )
					{
						//   
						::WSASetLastError( WSAETIMEDOUT );
						iRet =ICMPHLP_ERR_TIMEOUT;
						break;
					}
					//      
					else
						continue;
				}
				else
				{
					//  IP       20  60 
					//     IP    
					ICMPHeader *sICMPHdr =(ICMPHeader *)( (char *)sIPHdr + (sIPHdr->ucHLen*4) );

					//char t[123];
					//sprintf( t, "%d\n%d", sICMPHdr->usSeq, usSeq );
					//MessageBox( 0, t, 0, 0);

					if( sICMPHdr->usID !=usPacketID || sICMPHdr->ulIDEx !=::GetCurrentThreadId() || sICMPHdr->usSeq !=usSeq )
					{
						if( cAdvTimer.Elapsed( iBegin, cAdvTimer.Get() ) >=(double)_iRecvTimeout )
						{
							//   
							::WSASetLastError( WSAETIMEDOUT );
							iRet =ICMPHLP_ERR_TIMEOUT;
							break;
						}
						// "" ICMP 
						//  
						else
							continue;
					}
					else
					{
						if( sICMPHdr->ucType ==ICMP_ECHOREPLY )
						{
							//  ICMP 
							//         
							__int64 i64EndTm =cTimer.Get();
							dRecvTime =cTimer.Elapsed( sICMPHdr->i64Timestamp, i64EndTm );
							iRecvBytes =iRecv;
							iRecvTTL =sIPHdr->ucTTL;
							break;
						}
						else
						{
							//  ICMP 
							iRet =ICMPHLP_ERR_ICMPTYPE;
							break;
						}// else
					}// else
				}// else
			}// else
		}// while

		free(sIPHdr);
	}// else

	return iRet;
}